#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jan 14 11:07:35 2017

@author: A Ortiz
"""
#%reset
import tensorflow as tf
import numpy as np
import os
from tensorflow.examples.tutorials.mnist import input_data
import matplotlib.pyplot as plt 
import math

def plotNNFilter(units):
    filters = units.shape[3]
    plt.figure(1, figsize=(20,20))
    n_columns = 6
    n_rows = math.ceil(filters / n_columns) + 1
    for i in range(filters):
        plt.subplot(n_rows, n_columns, i+1)
        plt.title('Filter ' + str(i))
        plt.imshow(units[0,:,:,i], interpolation="nearest", cmap="gray")
        plt.axis('off')
    plt.show()    
        
# Carga datos
mnist = input_data.read_data_sets("../mnist", one_hot=True)

n_classes = 10
batch_size=100
n_iter=400
imshape=[28,28]

# X_batch son los datos, Y_batch las etiquetas onhot
#X_batch, Y_batch = mnist.train.next_batch(batch_size)

#Creamos sesion interactiva
sess = tf.InteractiveSession(config=tf.ConfigProto(intra_op_parallelism_threads=2))

# Placeholders para los tensores que almacenan los datos y las etoquetas
x = tf.placeholder(tf.float32, shape=[None, imshape[0], imshape[1],1]) # none to accept any batch size, 784 = 28x28
y_ = tf.placeholder(tf.float32, shape=[None, n_classes])

# Definimos las variables
# Pesos
def weight_variable(shape, name=None):
  initial = tf.truncated_normal(shape, stddev=0.1, dtype=tf.float32)
  return tf.Variable(initial, name=name)

# Bias
def bias_variable(shape, name=None):
  initial = tf.constant(0.1, shape=shape, dtype=tf.float32)
  return tf.Variable(initial, name=name)
  
# Definimos la funcion de convolucion: convolucion de x on W (kernel), stride=1 en las dos direcciones 
def conv2d(x, W, name=None):
  return tf.nn.conv2d(x, W, strides=[1, 1, 1, 1], padding='SAME', name=name)

# Definimos la funcion de pooling: pooling de x con un kernel de tamaño 2x2, stride=2 en las dos direcciones
# Esta capa reduce a 1/2 el tamaño !
def max_pool_2x2(x, name=None):
  return tf.nn.max_pool(x, ksize=[1, 2, 2, 1],
                        strides=[1, 2, 2, 1], padding='SAME', name=name)


# Primera capa
W_conv1 = weight_variable([5, 5, 1, 8], name='Wconv1')
b_conv1 = bias_variable([8], name='bconv1')

h_conv1 = tf.nn.relu(conv2d(x, W_conv1, name='relu1') + b_conv1)   # Activacion antes del pooling!
h_pool1 = max_pool_2x2(h_conv1, name='pool1')

# Segunda capa
W_conv2 = weight_variable([5, 5, 8, 16], name='Wconv2')
b_conv2 = bias_variable([16], name='bconv2')

h_conv2 = tf.nn.relu(conv2d(h_pool1, W_conv2, name='relu2') + b_conv2)
h_pool2 = max_pool_2x2(h_conv2, name='pool2')

# Tercera capa
W_conv3 = weight_variable([3, 3, 16, 32], name='Wconv3')
b_conv3 = bias_variable([32], name='bconv3')

h_conv3 = tf.nn.relu(conv2d(h_pool2, W_conv3, name='relu3') + b_conv3)
h_pool3 = max_pool_2x2(h_conv3, name='pool2')

# Tamaño final: 28x28 -- poolinx 2x2 --> 14x14 -- pooling 2x2 --> 7x7
finalsize = np.ceil(np.asarray(imshape)/8).astype(int)

# Capa Full-Connected
W_fc1 = weight_variable([finalsize[0] * finalsize[1] * 32, 4000], name='fc1')
b_fc1 = bias_variable([4000], name='bfc1')
#
h_pool3_flat = tf.reshape(h_pool3, [-1, finalsize[0] * finalsize[1] * 32], name='pool2f')
h_fc1 = tf.nn.relu(tf.matmul(h_pool3_flat, W_fc1) + b_fc1, name='fcrelu')   # Activacion ReLU
#
# Capa Dropout
keep_prob = tf.placeholder(tf.float32)
h_fc1_drop = tf.nn.dropout(h_fc1, keep_prob)
#
# Capa de salida
W_fc2 = weight_variable([4000, n_classes])
b_fc2 = bias_variable([n_classes])
#
y_conv = tf.matmul(h_fc1_drop, W_fc2) + b_fc2

# Funcion de perdida
cross_entropy = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits=y_conv, labels=y_))
# Optimizador
train_step = tf.train.AdamOptimizer(1e-4).minimize(cross_entropy)
# Definicion Accuracy
correct_prediction = tf.equal(tf.argmax(y_conv,1), tf.argmax(y_,1))
accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))

# Inicializa variables
sess.run(tf.global_variables_initializer())


accTraining=list()
accuracies=list()
loss=list()

for i in range(n_iter):
    x_batch_train, y_batch_train = mnist.train.next_batch(batch_size)
    # Las imagenes estan almacenadas como un vector de dim=784. Se convierte en tensores 2D 28x28 pixels  
    x_batch_train = np.reshape(x_batch_train, [-1,imshape[0], imshape[1], 1])
    train_step.run(feed_dict={x:x_batch_train, y_:y_batch_train, keep_prob: 0.5})
    trainacc = accuracy.eval(feed_dict={x:x_batch_train, y_: y_batch_train, keep_prob:1.0})
    accTraining.append(trainacc)
    ce=cross_entropy.eval(feed_dict={x:x_batch_train, y_:y_batch_train, keep_prob:1.0})
    if i % 50==0:
        print("Iter: %d, Trainin Accuracy %1.3f, loss=%1.2f"%(i, trainacc, ce))
    loss.append(ce)
    
# Evaluacion con test
accTest=list()
for i in range(10):
    x_batch_test,y_batch_test=mnist.test.next_batch(100)
    x_batch_test = np.reshape(x_batch_test, [-1,imshape[0], imshape[1], 1])
    testacc = accuracy.eval(feed_dict={x:x_batch_test, y_: y_batch_test, keep_prob:1.0})
    accTest.append(testacc)
    
    
# Evaluamos para guardar los pesos en un array (numpy)
wconv1_value = W_conv1.eval()
wconv2_value = W_conv2.eval()
bconv1_value = b_conv1.eval()
bconv2_value = b_conv2.eval()
# Cerramos la sesion 
sess.close()
print("Final accuracy %1.3f ± %1.3f"%(np.mean(accTest), np.std(accTest)))

################# Grafico loss / accuracy ###############
loss=np.asarray(loss)
accTraining=np.asarray(accTraining)
fig, ax1 = plt.subplots()
ax1.plot(range(len(loss)),loss,'b',label='Loss')
ax1.set_ylabel('Loss', color='b',fontsize=14)
ax1.set_xlabel('Iteration',fontsize=14)
ax1.tick_params('y', colors='b')
ax2 = ax1.twinx()
ax2.plot(range(len(accTraining)),accTraining,'r',label='Accuracy')
ax2.set_ylabel('Accuracy', color='r',fontsize=14)
ax2.tick_params('y', colors='r')
fig.tight_layout()
plt.show()

################ Visualizacion Filtros ####################
plt.figure(figsize=(6, 4))
# Primera capa (8 filtros)
for i in range(8):
    plt.subplot(2,4,i+1)
    plt.imshow(wconv1_value[:,:,0,i],cmap='gray')
    plt.draw()
    plt.pause(0.00001)
    plt.axis('off')
plt.show()

plt.figure(figsize=(6, 4))
# Segunda Capa (16 filtros)
for i in range(16):
    plt.subplot(4,4,i+1)
    plt.imshow(wconv2_value[:,:,0,i],cmap='gray')
    plt.draw()
    plt.pause(0.00001)
    plt.axis('off')
plt.show()
    
############## Visualizacion activaciones #################
# Visualizacion de las activaciones de la primera y segunda capa
# Los filtros estan almacenados en W_conv1 y W_conv2 / b_conv1, b_conv2
# Cargamos una imagen (estimulo)
print('Entrada')
img=mnist.test.images[1]
plt.imshow(img.reshape(28,28),cmap='gray')
plt.axis('off')
plt.show()
x_img = np.reshape(img, [-1,28, 28, 1])
    
# Creamos la sesion
sess = tf.InteractiveSession(config=tf.ConfigProto(intra_op_parallelism_threads=2))
units1=h_conv1.eval(feed_dict={x:x_img, W_conv1:wconv1_value, b_conv1:bconv1_value})
print('Activacion Capa 1')
plotNNFilter(units1)
pool1=h_pool1.eval(feed_dict={h_conv1:units1})      
units2=h_pool2.eval(feed_dict={h_pool1:pool1, W_conv2:wconv2_value, b_conv2:bconv2_value})
print('Activacion Capa 2')
plotNNFilter(units2)
sess.close()



