#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug  8 11:47:15 2017

@author: andres
"""

# Perceptron multicapa (MLP)


import tensorflow as tf
import matplotlib.pyplot as plt
import numpy as np
from tensorflow.examples.tutorials.mnist import input_data

# Parametros
learning_rate=0.001
train_iters=20
batch_size=100  # Tamaño del batch: entrenamos de 100 en 100 imagenes

n_input=784 # Numero de neuronas capa entrada. Las imagenes mnist son de 28x28 = 784 pixels
n_h1=256 # Numero de neuronas capa oculta 1
n_h2=256 # Numero de neuronas capa oculta 2
n_clases=10 # Numero de neuronas capa salida

# Lectura de los datos
print('Cargando datos...')
mnist = input_data.read_data_sets("../mnist", one_hot=True) # One hot -> vector binario de clases

# Definicion de arrays para realizar graficos
avgcost_set=[]
avgacc_train=[]
avgacc_test=[]
iter_set=[]

# Definicion de placeholders para almacenar los datos (imagenes y etiquetas)
x=tf.placeholder(tf.float32,[None,n_input])  # Capa Entrada. Con None se permite dimensionado dinamico
y=tf.placeholder(tf.float32,[None,n_clases]) # Capa Salida

# Definiciones capa oculta 1. 
h1=tf.Variable(tf.random_normal([n_input,n_h1])) # Pesos 784x n_h1
b1=tf.Variable(tf.random_normal([n_h1])) # Bias n_h1
#l1=tf.nn.sigmoid(tf.add(tf.matmul(x,h1),b1))  # Activacion sigmoidal capa 1, xW+b
l1=tf.nn.relu(tf.add(tf.matmul(x,h1),b1))  # Activacion ReLU capa 1, xW+b

# Definiciones capa salida
output=tf.Variable(tf.random_normal([n_h1,n_clases]))  # Pesos n_h2 x n_clases
bo=tf.Variable(tf.random_normal([n_clases])) # Bias n_clases
lo=tf.matmul(l1,output)+bo # Activacion 

# Definicion funcion de costo. Calculo de la entropia cruzada en la capa softmax
# media_sobre_todas_las_muestras (softmax -> cross_entropy) 
cost=tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(labels=y,logits=lo))

# Optimizador Adam
optim=tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(cost)

# Definicion del inicializador
init=tf.global_variables_initializer()

# Clase con la maxima probabilidad softmax
pred=tf.equal(tf.argmax(lo,1),tf.argmax(y,1))
# Definicion Accuracy
accuracy=tf.reduce_mean(tf.cast(pred,tf.float32))

# Definicion de la sesion
with tf.Session() as sess:
    sess.run(init)
    for iter in range(train_iters):
        avg_cost=0
        # Entrenamos de forma incremental usando batches
        total_batch=int(mnist.train.num_examples/batch_size) # Numero de batches
        for i in range(total_batch):
            # Cargamos un batch
            batch_x, batch_y=mnist.train.next_batch(batch_size)
            # Entrenamos con los datos del batch
            sess.run(optim,feed_dict={x:batch_x, y:batch_y})
            # Calculo del coste medio para el batch
            avg_cost+=sess.run(cost,feed_dict={x:batch_x, y:batch_y})/total_batch
            
        # Listas para almacenar el coste (para graficos)
        avgcost_set.append(avg_cost)
        iter_set.append(iter+1)
        avgacc_test.append(1-accuracy.eval({x:mnist.test.images,y:mnist.test.labels}))
        avgacc_train.append(1-accuracy.eval({x:mnist.train.images,y:mnist.train.labels}))

        
        print("Iteracion %d, Coste medio=%1.2f"%(iter,avg_cost))
            
    # Test: Calculo de accuracy
    
    # Calculo del accuracy medio
    print("Precision media del modelo:%1.2f"%(accuracy.eval({x:mnist.test.images,y:mnist.test.labels})))
    
    
# Graficos
fig=plt.figure()
ax = fig.add_subplot(111)
plt.plot(iter_set,avgcost_set,'bo-',label='MLP')
plt.xticks(iter_set,np.arange(0,train_iters))
plt.xlabel('Iteracion',fontsize=14)
plt.ylabel('Coste',fontsize=14)
plt.legend()
plt.show()

fig=plt.figure()
x = fig.add_subplot(111)
plt.plot(iter_set,avgacc_train,'bo-',label='Entrenamiento MLP')
plt.plot(iter_set,avgacc_test,'ro-',label='Test MLP')
plt.xticks(iter_set,np.arange(0,train_iters))
plt.xlabel('Iteracion',fontsize=14)
plt.ylabel('Accuracy',fontsize=14)
plt.legend()
plt.show()











